# Generate a loop kernel from that cell integral. This will map the UFL IR
# to the loopy IR.
from ufl.algorithms import MultiFunction
from pymbolic.primitives import Variable, Subscript, Sum, Product
from loopy.kernel.data import ExpressionInstruction, CInstruction
import loopy
import numpy

# For now, import all implemented preambles, restrict later
from dune.perftool.pdelab_preambles import *
from dune.perftool.restriction import Restriction

class UFLVisitor(MultiFunction):
    def __call__(self, o):
        # Make this multifunction stateful: Store information similar to uflacs' modified terminal
        self.grad = False
        self.reference_grad = False
        self.index = None
        self.restriction = Restriction.NONE

        # Have a short cut for the bases class call operator
        self.call = lambda *a : MultiFunction.__call__(self, *a)

        # Have a list of argument indices that this term depends on.
        self.argument_indices = []

        loopyexpr = self.call(o)

        # TODO no jacobian support yet!
        assert len(self.argument_indices) == 1

        assignee = Subscript(Variable("r"), Variable(self.argument_indices[0]))
        loopyexpr = Sum((assignee, Product((Variable(quadrature_factor()), loopyexpr))))

        instruction = ExpressionInstruction(assignee=assignee, expression=loopyexpr)
        accumulation_instruction(instruction)

    def grad(self, o):
        assert(len(o.operands()) == 1)

        self.grad = True
        ret = self.call(o.operands()[0])
        self.grad = False
        return ret

    def argument(self, o):
        assert(len(o.operands()) == 0)

        # We do need an argument loop domain for this argument
        argindex = argument_iname(o)

        # Generate the variable name for the test function
        name = test_function_name(o, self.grad)
        index = index_name(self.index)

        # Register that index as argument loop index for later
        self.argument_indices.append(argindex)

        return Subscript(Variable(name), Variable(index))

    def coefficient(self, o):
        assert(len(o.operands()) == 0)

        name = trial_function_name(o, self.grad)
        index = index_name(self.index)
        return Subscript(Variable(name), Variable(index))

    def product(self, o):
        return Product(tuple(self.call(op) for op in o.operands()))

    def multi_index(self, o):
        # I don't think we should ever find a multi index, because its father should take care of it.
        raise ValueError

    def indexed(self, o):
        # TODO in the long run, this is a stack of indices.
        self.index = o.operands()[1]
        ret = self.call(o.operands()[0])
        self.index = None
        return ret


class TopSumSeparation(MultiFunction):
    """ A multifunction that separates the toplevel sum """
    def __init__(self):
        MultiFunction.__init__(self)
        self.visitor = UFLVisitor()

    def expr(self, o):
        print "Call TopSumSeparation.expr with {}".format(o)
        self.visitor(o)

    def sum(self, o):
        for op in o.operands():
            self(op)

    def index_sum(self, o):
        print "Call TopSumSeparation.index_sum with {}".format(o)
        dimension_iname(o.operands()[1])
        self(o.operands()[0])


def transform_expression(expr):
    return TopSumSeparation()(expr)
